"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, cast

import wavelink

from .enums import DiscordVoiceCloseType
from .filters import Filters
from .tracks import Playable

if TYPE_CHECKING:
    from .node import Node
    from .player import Player
    from .types.filters import *
    from .types.response import *
    from .types.state import PlayerState
    from .types.stats import CPUStats, FrameStats, MemoryStats
    from .types.websocket import StatsOP, TrackExceptionPayload


__all__ = (
    "TrackStartEventPayload",
    "TrackEndEventPayload",
    "TrackExceptionEventPayload",
    "TrackStuckEventPayload",
    "WebsocketClosedEventPayload",
    "PlayerUpdateEventPayload",
    "StatsEventPayload",
    "NodeReadyEventPayload",
    "StatsEventMemory",
    "StatsEventCPU",
    "StatsEventFrames",
    "StatsResponsePayload",
    "GitResponsePayload",
    "VersionResponsePayload",
    "PluginResponsePayload",
    "InfoResponsePayload",
    "PlayerStatePayload",
    "VoiceStatePayload",
    "PlayerResponsePayload",
    "ExtraEventPayload",
)


class NodeReadyEventPayload:
    """Payload received in the :func:`on_wavelink_node_ready` event.

    Attributes
    ----------
    node: :class:`~wavelink.Node`
        The node that has connected or reconnected.
    resumed: bool
        Whether this node was successfully resumed.
    session_id: str
        The session ID associated with this node.
    """

    def __init__(self, node: Node, resumed: bool, session_id: str) -> None:
        self.node = node
        self.resumed = resumed
        self.session_id = session_id


class TrackStartEventPayload:
    """Payload received in the :func:`on_wavelink_track_start` event.

    Attributes
    ----------
    player: :class:`~wavelink.Player` | None
        The player associated with this event. Could be None.
    track: :class:`~wavelink.Playable`
        The track received from Lavalink regarding this event.
    original: :class:`~wavelink.Playable` | None
        The original track associated this event. E.g. the track that was passed to :meth:`~wavelink.Player.play` or
        inserted into the queue, with all your additional attributes assigned. Could be ``None``.
    """

    def __init__(self, player: Player | None, track: Playable) -> None:
        self.player = player
        self.track = track
        self.original: Playable | None = None

        if player:
            self.original = player._original


class TrackEndEventPayload:
    """Payload received in the :func:`on_wavelink_track_end` event.

    Attributes
    ----------
    player: :class:`~wavelink.Player` | None
        The player associated with this event. Could be None.
    track: :class:`~wavelink.Playable`
        The track received from Lavalink regarding this event.
    reason: str
        The reason Lavalink ended this track.
    original: :class:`~wavelink.Playable` | None
        The original track associated this event. E.g. the track that was passed to :meth:`~wavelink.Player.play` or
        inserted into the queue, with all your additional attributes assigned. Could be ``None``.
    """

    def __init__(self, player: Player | None, track: Playable, reason: str) -> None:
        self.player = player
        self.track = track
        self.reason = reason
        self.original: Playable | None = None

        if player:
            self.original = player._previous


class TrackExceptionEventPayload:
    """Payload received in the :func:`on_wavelink_track_exception` event.

    Attributes
    ----------
    player: :class:`~wavelink.Player` | None
        The player associated with this event. Could be None.
    track: :class:`~wavelink.Playable`
        The track received from Lavalink regarding this event.
    exception: TrackExceptionPayload
        The exception data received via Lavalink.
    """

    def __init__(self, player: Player | None, track: Playable, exception: TrackExceptionPayload) -> None:
        self.player = cast(wavelink.Player, player)
        self.track = track
        self.exception = exception


class TrackStuckEventPayload:
    """Payload received in the :func:`on_wavelink_track_stuck` event.

    Attributes
    ----------
    player: :class:`~wavelink.Player` | None
        The player associated with this event. Could be None.
    track: :class:`~wavelink.Playable`
        The track received from Lavalink regarding this event.
    threshold: int
        The Lavalink threshold associated with this event.
    """

    def __init__(self, player: Player | None, track: Playable, threshold: int) -> None:
        self.player = cast(wavelink.Player, player)
        self.track = track
        self.threshold = threshold


class WebsocketClosedEventPayload:
    """Payload received in the :func:`on_wavelink_websocket_closed` event.

    Attributes
    ----------
    player: :class:`~wavelink.Player` | None
        The player associated with this event. Could be None.
    code: :class:`wavelink.DiscordVoiceCloseType`
        The close code enum value.
    reason: str
        The reason the websocket was closed.
    by_remote: bool
        ``True`` if discord closed the websocket. ``False`` otherwise.
    """

    def __init__(self, player: Player | None, code: int, reason: str, by_remote: bool) -> None:
        self.player = player
        self.code: DiscordVoiceCloseType = DiscordVoiceCloseType(code)
        self.reason = reason
        self.by_remote = by_remote


class PlayerUpdateEventPayload:
    """Payload received in the :func:`on_wavelink_player_update` event.

    Attributes
    ----------
    player: :class:`~wavelink.Player` | None
        The player associated with this event. Could be None.
    time: int
        Unix timestamp in milliseconds, when this event fired.
    position: int
        The position of the currently playing track in milliseconds.
    connected: bool
        Whether Lavalink is connected to the voice gateway.
    ping: int
        The ping of the node to the Discord voice server in milliseconds (-1 if not connected).
    """

    def __init__(self, player: Player | None, state: PlayerState) -> None:
        self.player = cast(wavelink.Player, player)
        self.time: int = state["time"]
        self.position: int = state["position"]
        self.connected: bool = state["connected"]
        self.ping: int = state["ping"]


class StatsEventMemory:
    """Represents Memory Stats.

    Attributes
    ----------
    free: int
        The amount of free memory in bytes.
    used: int
        The amount of used memory in bytes.
    allocated: int
        The amount of allocated memory in bytes.
    reservable: int
        The amount of reservable memory in bytes.
    """

    def __init__(self, data: MemoryStats) -> None:
        self.free: int = data["free"]
        self.used: int = data["used"]
        self.allocated: int = data["allocated"]
        self.reservable: int = data["reservable"]


class StatsEventCPU:
    """Represents CPU Stats.

    Attributes
    ----------
    cores: int
        The number of CPU cores available on the node.
    system_load: float
        The system load of the node.
    lavalink_load: float
        The load of Lavalink on the node.
    """

    def __init__(self, data: CPUStats) -> None:
        self.cores: int = data["cores"]
        self.system_load: float = data["systemLoad"]
        self.lavalink_load: float = data["lavalinkLoad"]


class StatsEventFrames:
    """Represents Frame Stats.

    Attributes
    ----------
    sent: int
        The amount of frames sent to Discord.
    nulled: int
        The amount of frames that were nulled.
    deficit: int
        The difference between sent frames and the expected amount of frames.
    """

    def __init__(self, data: FrameStats) -> None:
        self.sent: int = data["sent"]
        self.nulled: int = data["nulled"]
        self.deficit: int = data["deficit"]


class StatsEventPayload:
    """Payload received in the :func:`on_wavelink_stats_update` event.

    Attributes
    ----------
    players: int
        The amount of players connected to the node (Lavalink).
    playing: int
        The amount of players playing a track.
    uptime: int
        The uptime of the node in milliseconds.
    memory: :class:`wavelink.StatsEventMemory`
        See Also: :class:`wavelink.StatsEventMemory`
    cpu: :class:`wavelink.StatsEventCPU`
        See Also: :class:`wavelink.StatsEventCPU`
    frames: :class:`wavelink.StatsEventFrames` | None
        See Also: :class:`wavelink.StatsEventFrames`. This could be ``None``.
    """

    def __init__(self, data: StatsOP) -> None:
        self.players: int = data["players"]
        self.playing: int = data["playingPlayers"]
        self.uptime: int = data["uptime"]

        self.memory: StatsEventMemory = StatsEventMemory(data=data["memory"])
        self.cpu: StatsEventCPU = StatsEventCPU(data=data["cpu"])
        self.frames: StatsEventFrames | None = None

        if frames := data.get("frameStats", None):
            self.frames = StatsEventFrames(frames)


class StatsResponsePayload:
    """Payload received when using :meth:`~wavelink.Node.fetch_stats`

    Attributes
    ----------
    players: int
        The amount of players connected to the node (Lavalink).
    playing: int
        The amount of players playing a track.
    uptime: int
        The uptime of the node in milliseconds.
    memory: :class:`wavelink.StatsEventMemory`
        See Also: :class:`wavelink.StatsEventMemory`
    cpu: :class:`wavelink.StatsEventCPU`
        See Also: :class:`wavelink.StatsEventCPU`
    frames: :class:`wavelink.StatsEventFrames` | None
        See Also: :class:`wavelink.StatsEventFrames`. This could be ``None``.
    """

    def __init__(self, data: StatsResponse) -> None:
        self.players: int = data["players"]
        self.playing: int = data["playingPlayers"]
        self.uptime: int = data["uptime"]

        self.memory: StatsEventMemory = StatsEventMemory(data=data["memory"])
        self.cpu: StatsEventCPU = StatsEventCPU(data=data["cpu"])
        self.frames: StatsEventFrames | None = None

        if frames := data.get("frameStats", None):
            self.frames = StatsEventFrames(frames)


class PlayerStatePayload:
    """Represents the PlayerState information received via :meth:`~wavelink.Node.fetch_player_info` or
    :meth:`~wavelink.Node.fetch_players`

    Attributes
    ----------
    time: int
        Unix timestamp in milliseconds received from Lavalink.
    position: int
        The position of the track in milliseconds received from Lavalink.
    connected: bool
        Whether Lavalink is connected to the voice gateway.
    ping: int
        The ping of the node to the Discord voice server in milliseconds (-1 if not connected).
    """

    def __init__(self, data: PlayerState) -> None:
        self.time: int = data["time"]
        self.position: int = data["position"]
        self.connected: bool = data["connected"]
        self.ping: int = data["ping"]


class VoiceStatePayload:
    """Represents the VoiceState information received via :meth:`~wavelink.Node.fetch_player_info` or
    :meth:`~wavelink.Node.fetch_players`. This is the voice state information received via Discord and sent to your
    Lavalink node.

    Attributes
    ----------
    token: str | None
        The Discord voice token authenticated with. This is not the same as your bots token. Could be ``None``.
    endpoint: str | None
        The Discord voice endpoint connected to. Could be ``None``.
    session_id: str | None
        The Discord voice session ID autheticated with. Could be ``None``.
    """

    def __init__(self, data: VoiceStateResponse) -> None:
        self.token: str | None = data.get("token")
        self.endpoint: str | None = data.get("endpoint")
        self.session_id: str | None = data.get("sessionId")


class PlayerResponsePayload:
    """Payload received when using :meth:`~wavelink.Node.fetch_player_info` or :meth:`~wavelink.Node.fetch_players`

    Attributes
    ----------
    guild_id: int
        The guild ID as an int that this player is connected to.
    track: :class:`wavelink.Playable` | None
        The current track playing on Lavalink. Could be ``None`` if no track is playing.
    volume: int
        The current volume of the player.
    paused: bool
        A bool indicating whether the player is paused.
    state: :class:`wavelink.PlayerStatePayload`
        The current state of the player. See: :class:`wavelink.PlayerStatePayload`.
    voice_state: :class:`wavelink.VoiceStatePayload`
        The voice state infomration received via Discord and sent to Lavalink. See: :class:`wavelink.VoiceStatePayload`.
    filters: :class:`wavelink.Filters`
        The :class:`wavelink.Filters` currently associated with this player.
    """

    def __init__(self, data: PlayerResponse) -> None:
        self.guild_id: int = int(data["guildId"])
        self.track: Playable | None = None

        if track := data.get("track"):
            self.track = Playable(track)

        self.volume: int = data["volume"]
        self.paused: bool = data["paused"]
        self.state: PlayerStatePayload = PlayerStatePayload(data["state"])
        self.voice_state: VoiceStatePayload = VoiceStatePayload(data["voice"])
        self.filters: Filters = Filters(data=data["filters"])


class GitResponsePayload:
    """Represents Git information received via :meth:`wavelink.Node.fetch_info`

    Attributes
    ----------
    branch: str
        The branch this Lavalink server was built on.
    commit: str
        The commit this Lavalink server was built on.
    commit_time: :class:`datetime.datetime`
        The timestamp for when the commit was created.
    """

    def __init__(self, data: GitPayload) -> None:
        self.branch: str = data["branch"]
        self.commit: str = data["commit"]
        self.commit_time: datetime.datetime = datetime.datetime.fromtimestamp(
            data["commitTime"] / 1000, tz=datetime.timezone.utc
        )


class VersionResponsePayload:
    """Represents Version information received via :meth:`wavelink.Node.fetch_info`

    Attributes
    ----------
    semver: str
        The full version string of this Lavalink server.
    major: int
        The major version of this Lavalink server.
    minor: int
        The minor version of this Lavalink server.
    patch: int
        The patch version of this Lavalink server.
    pre_release: str
        The pre-release version according to semver as a ``.`` separated list of identifiers.
    build: str | None
        The build metadata according to semver as a ``.`` separated list of identifiers. Could be ``None``.
    """

    def __init__(self, data: VersionPayload) -> None:
        self.semver: str = data["semver"]
        self.major: int = data["major"]
        self.minor: int = data["minor"]
        self.patch: int = data["patch"]
        self.pre_release: str | None = data.get("preRelease")
        self.build: str | None = data.get("build")


class PluginResponsePayload:
    """Represents Plugin information received via :meth:`wavelink.Node.fetch_info`

    Attributes
    ----------
    name: str
        The plugin name.
    version: str
        The plugin version.
    """

    def __init__(self, data: PluginPayload) -> None:
        self.name: str = data["name"]
        self.version: str = data["version"]


class InfoResponsePayload:
    """Payload received when using :meth:`~wavelink.Node.fetch_info`

    Attributes
    ----------
    version: :class:`VersionResponsePayload`
        The version info payload for this Lavalink node in the :class:`VersionResponsePayload` object.
    build_time: :class:`datetime.datetime`
        The timestamp when this Lavalink jar was built.
    git: :class:`GitResponsePayload`
        The git info payload for this Lavalink node in the :class:`GitResponsePayload` object.
    jvm: str
        The JVM version this Lavalink node runs on.
    lavaplayer: str
        The Lavaplayer version being used by this Lavalink node.
    source_managers: list[str]
        The enabled source managers for this node.
    filters: list[str]
        The enabled filters for this node.
    plugins: list[:class:`PluginResponsePayload`]
        The enabled plugins for this node.
    """

    def __init__(self, data: InfoResponse) -> None:
        self.version: VersionResponsePayload = VersionResponsePayload(data["version"])
        self.build_time: datetime.datetime = datetime.datetime.fromtimestamp(
            data["buildTime"] / 1000, tz=datetime.timezone.utc
        )
        self.git: GitResponsePayload = GitResponsePayload(data["git"])
        self.jvm: str = data["jvm"]
        self.lavaplayer: str = data["lavaplayer"]
        self.source_managers: list[str] = data["sourceManagers"]
        self.filters: list[str] = data["filters"]
        self.plugins: list[PluginResponsePayload] = [PluginResponsePayload(p) for p in data["plugins"]]


class ExtraEventPayload:
    """Payload received in the :func:`on_wavelink_extra_event` event.

    This payload is created when an ``Unknown`` and ``Unhandled`` event is received from Lavalink, most likely via
    a plugin.

    .. note::

        See the appropriate documentation of the plugin for the data sent with these events.


    Attributes
    ----------
    node: :class:`~wavelink.Node`
        The node that the event pertains to.
    player: :class:`~wavelink.Player` | None
        The player associated with this event. Could be None.
    data: dict[str, Any]
        The raw data sent from Lavalink for this event.


    .. versionadded:: 3.1.0
    """

    def __init__(self, *, node: Node, player: Player | None, data: dict[str, Any]) -> None:
        self.node = node
        self.player = player
        self.data = data



"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypeAlias, TypedDict

if TYPE_CHECKING:
    from typing_extensions import NotRequired

    from .state import PlayerState
    from .stats import CPUStats, FrameStats, MemoryStats
    from .tracks import TrackPayload


class TrackExceptionPayload(TypedDict):
    message: NotRequired[str]
    severity: str
    cause: str


class ReadyOP(TypedDict):
    op: Literal["ready"]
    resumed: bool
    sessionId: str


class PlayerUpdateOP(TypedDict):
    op: Literal["playerUpdate"]
    guildId: str
    state: PlayerState


class StatsOP(TypedDict):
    op: Literal["stats"]
    players: int
    playingPlayers: int
    uptime: int
    memory: MemoryStats
    cpu: CPUStats
    frameStats: FrameStats


class TrackStartEvent(TypedDict):
    op: Literal["event"]
    guildId: str
    type: Literal["TrackStartEvent"]
    track: TrackPayload


class TrackEndEvent(TypedDict):
    op: Literal["event"]
    guildId: str
    type: Literal["TrackEndEvent"]
    track: TrackPayload
    reason: str


class TrackExceptionEvent(TypedDict):
    op: Literal["event"]
    guildId: str
    type: Literal["TrackExceptionEvent"]
    track: TrackPayload
    exception: TrackExceptionPayload


class TrackStuckEvent(TypedDict):
    op: Literal["event"]
    guildId: str
    type: Literal["TrackStuckEvent"]
    track: TrackPayload
    thresholdMs: int


class WebsocketClosedEvent(TypedDict):
    op: Literal["event"]
    guildId: str
    type: Literal["WebSocketClosedEvent"]
    code: int
    reason: str
    byRemote: bool


WebsocketOP: TypeAlias = (
    ReadyOP
    | PlayerUpdateOP
    | StatsOP
    | TrackStartEvent
    | TrackEndEvent
    | TrackExceptionEvent
    | TrackStuckEvent
    | WebsocketClosedEvent
)



"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import asyncio
import random
from collections import deque
from collections.abc import Iterable, Iterator
from typing import SupportsIndex, TypeGuard, overload

from .enums import QueueMode
from .exceptions import QueueEmpty
from .tracks import Playable, Playlist

__all__ = ("Queue",)


class Queue:
    """The default custom wavelink Queue designed specifically for :class:`wavelink.Player`.

    .. note::

        :class:`~wavelink.Player` implements this queue by default.
        You can access it via :attr:`wavelink.Player.queue`.

    .. container:: operations

        .. describe:: str(queue)

            A string representation of this queue.

        .. describe:: repr(queue)

            The official string representation of this queue.

        .. describe:: if queue

            Bool check whether this queue has items or not.

        .. describe:: queue(track)

            Put a track in the queue.

        .. describe:: len(queue)

            The amount of tracks in the queue.

        .. describe:: queue[1]

            Peek at an item in the queue. Does not change the queue.

        .. describe:: for item in queue

            Iterate over the queue.

        .. describe:: if item in queue

            Check whether a specific track is in the queue.

        .. describe:: queue[1] = track

            Set a track in the queue at a specific index.

        .. describe:: del queue[1]

            Delete a track from the queue at a specific index.

        .. describe:: reversed(queue)

            Return a reversed iterator of the queue.

    Attributes
    ----------
    history: :class:`wavelink.Queue`
        A queue of tracks that have been added to history. Tracks are added to history when they are played.
    """

    def __init__(self, *, history: bool = True) -> None:
        self._items: list[Playable] = []

        self._history: Queue | None = Queue(history=False) if history else None
        self._mode: QueueMode = QueueMode.normal
        self._loaded: Playable | None = None

        self._waiters: deque[asyncio.Future[None]] = deque()
        self._lock = asyncio.Lock()

    @property
    def mode(self) -> QueueMode:
        """Property which returns a :class:`~wavelink.QueueMode` indicating which mode the
        :class:`~wavelink.Queue` is in.

        This property can be set with any :class:`~wavelink.QueueMode`.


        .. versionadded:: 3.0.0
        """
        return self._mode

    @mode.setter
    def mode(self, value: QueueMode) -> None:
        self._mode = value

    @property
    def history(self) -> Queue | None:
        return self._history

    @property
    def count(self) -> int:
        """The queue member count.

        Returns
        -------
        int
            The amount of tracks in the queue.


        .. versionadded:: 3.2.0
        """

        return len(self)

    @property
    def is_empty(self) -> bool:
        """Whether the queue has no members.

        Returns
        -------
        bool
            Whether the queue is empty.


        .. versionadded:: 3.2.0
        """

        return not bool(self)

    def __str__(self) -> str:
        joined: str = ", ".join([f'"{p}"' for p in self])
        return f"Queue([{joined}])"

    def __repr__(self) -> str:
        return f"Queue(items={len(self)}, history={self.history!r})"

    def __call__(self, item: Playable) -> None:
        self.put(item)

    def __bool__(self) -> bool:
        return bool(self._items)

    @overload
    def __getitem__(self, __index: SupportsIndex, /) -> Playable:
        ...

    @overload
    def __getitem__(self, __index: slice, /) -> list[Playable]:
        ...

    def __getitem__(self, __index: SupportsIndex | slice, /) -> Playable | list[Playable]:
        return self._items[__index]

    def __setitem__(self, __index: SupportsIndex, __value: Playable, /) -> None:
        self._check_compatibility(__value)
        self._items[__index] = __value
        self._wakeup_next()

    def __delitem__(self, __index: int | slice, /) -> None:
        del self._items[__index]

    def __contains__(self, __other: Playable) -> bool:
        return __other in self._items

    def __len__(self) -> int:
        return len(self._items)

    def __reversed__(self) -> Iterator[Playable]:
        return reversed(self._items)

    def __iter__(self) -> Iterator[Playable]:
        return iter(self._items)

    def _wakeup_next(self) -> None:
        while self._waiters:
            waiter = self._waiters.popleft()

            if not waiter.done():
                waiter.set_result(None)
                break

    @staticmethod
    def _check_compatibility(item: object) -> TypeGuard[Playable]:
        if not isinstance(item, Playable):
            raise TypeError("This queue is restricted to Playable objects.")
        return True

    @classmethod
    def _check_atomic(cls, item: Iterable[object]) -> TypeGuard[Iterable[Playable]]:
        for track in item:
            cls._check_compatibility(track)
        return True

    def get(self) -> Playable:
        """Retrieve a track from the left side of the queue. E.g. the first.

        This method does not block.

        .. warning::

            Due to the way the queue loop works, this method will return the same track if the queue is in loop mode.
            You can use :meth:`wavelink.Player.skip` with ``force=True`` to skip the current track.

            Do **NOT** use this method to remove tracks from the queue, use either:

            - ``del queue[index]``
            - :meth:`wavelink.Queue.remove`
            - :meth:`wavelink.Queue.delete`


        Returns
        -------
        :class:`wavelink.Playable`
            The track retrieved from the queue.

        Raises
        ------
        QueueEmpty
            The queue was empty when retrieving a track.
        """

        if self.mode is QueueMode.loop and self._loaded:
            return self._loaded

        if self.mode is QueueMode.loop_all and not self:
            assert self.history is not None

            self._items.extend(self.history._items)
            self.history.clear()

        if not self:
            raise QueueEmpty("There are no items currently in this queue.")

        track: Playable = self._items.pop(0)
        self._loaded = track

        return track

    def get_at(self, index: int, /) -> Playable:
        """Retrieve a track from the queue at a given index.

        .. warning::

            Due to the way the queue loop works, this method will load the retrieved track for looping.

            Do **NOT** use this method to remove tracks from the queue, use either:

            - ``del queue[index]``
            - :meth:`wavelink.Queue.remove`
            - :meth:`wavelink.Queue.delete`

        Parameters
        ----------
        index: int
            The index of the track to get.

        Returns
        -------
        :class:`wavelink.Playable`
            The track retrieved from the queue.

        Raises
        ------
        QueueEmpty
            The queue was empty when retrieving a track.
        IndexError
            The index was out of range for the current queue.


        .. versionadded:: 3.2.0
        """

        if not self:
            raise QueueEmpty("There are no items currently in this queue.")

        track: Playable = self._items.pop(index)
        self._loaded = track

        return track

    def put_at(self, index: int, value: Playable, /) -> None:
        """Put a track into the queue at a given index.

        .. note::

            This method doesn't replace the track at the index but rather inserts one there, similar to a list.

        Parameters
        ----------
        index: int
            The index to put the track at.
        value: :class:`wavelink.Playable`
            The track to put.

        Raises
        ------
        TypeError
            The track was not a :class:`wavelink.Playable`.


        .. versionadded:: 3.2.0
        """
        self._check_compatibility(value)
        self._items.insert(index, value)
        self._wakeup_next()

    async def get_wait(self) -> Playable:
        """This method returns the first :class:`wavelink.Playable` if one is present or
        waits indefinitely until one is.

        This method is asynchronous.

        Returns
        -------
        :class:`wavelink.Playable`
            The track retrieved from the queue.
        """

        while not self:
            loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
            waiter: asyncio.Future[None] = loop.create_future()

            self._waiters.append(waiter)

            try:
                await waiter
            except:  # noqa
                waiter.cancel()

                try:
                    self._waiters.remove(waiter)
                except ValueError:  # pragma: no branch
                    pass

                if self and not waiter.cancelled():  # pragma: no cover
                    # something went wrong with this waiter, move on to next
                    self._wakeup_next()
                raise

        return self.get()

    def put(self, item: list[Playable] | Playable | Playlist, /, *, atomic: bool = True) -> int:
        """Put an item into the end of the queue.

        Accepts a :class:`wavelink.Playable`, :class:`wavelink.Playlist` or list[:class:`wavelink.Playble`].

        Parameters
        ----------
        item: :class:`wavelink.Playable` | :class:`wavelink.Playlist` | list[:class:`wavelink.Playble`]
            The item to enter into the queue.
        atomic: bool
            Whether the items should be inserted atomically. If set to ``True`` this method won't enter any tracks if
            it encounters an error. Defaults to ``True``.

        Returns
        -------
        int
            The number of tracks added to the queue.
        """

        added = 0

        if isinstance(item, Iterable):
            if atomic:
                self._check_atomic(item)
                self._items.extend(item)
                added = len(item)
            else:

                def try_compatibility(track: object) -> bool:
                    try:
                        return self._check_compatibility(track)
                    except TypeError:
                        return False

                passing_items = [track for track in item if try_compatibility(track)]
                self._items.extend(passing_items)
                added = len(passing_items)
        else:
            self._check_compatibility(item)
            self._items.append(item)
            added = 1

        self._wakeup_next()
        return added

    async def put_wait(self, item: list[Playable] | Playable | Playlist, /, *, atomic: bool = True) -> int:
        """Put an item or items into the end of the queue asynchronously.

        Accepts a :class:`wavelink.Playable` or :class:`wavelink.Playlist` or list[:class:`wavelink.Playable`].

        .. note::

            This method implements a lock to preserve insert order.

        Parameters
        ----------
        item: :class:`wavelink.Playable` | :class:`wavelink.Playlist` | list[:class:`wavelink.Playable`]
            The item or items to enter into the queue.
        atomic: bool
            Whether the items should be inserted atomically. If set to ``True`` this method won't enter any tracks if
            it encounters an error. Defaults to ``True``.

        Returns
        -------
        int
            The number of tracks added to the queue.
        """

        added: int = 0

        async with self._lock:
            if isinstance(item, Iterable):
                if atomic:
                    self._check_atomic(item)
                    self._items.extend(item)
                    return len(item)

                for track in item:
                    try:
                        self._check_compatibility(track)
                    except TypeError:
                        pass
                    else:
                        self._items.append(track)
                        added += 1

                    await asyncio.sleep(0)

            else:
                self._check_compatibility(item)
                self._items.append(item)
                added += 1
                await asyncio.sleep(0)

        self._wakeup_next()
        return added

    def delete(self, index: int, /) -> None:
        """Method to delete an item in the queue by index.

        Raises
        ------
        IndexError
            No track exists at this index.

        Examples
        --------

        .. code:: python3

            # Deletes the track at index 1 (The second track).
            queue.delete(1)


        .. versionchanged:: 3.2.0

            The method is no longer a coroutine.
        """

        del self._items[index]

    def peek(self, index: int = 0, /) -> Playable:
        """Method to peek at an item in the queue by index.

        .. note::

            This does not change the queue or remove the item.

        Parameters
        ----------
        index: int
            The index to peek at. Defaults to ``0`` which is the next item in the queue.

        Returns
        -------
        :class:`wavelink.Playable`
            The track at the given index.

        Raises
        ------
        QueueEmpty
            There are no items currently in this queue.
        IndexError
            No track exists at the given index.


        .. versionadded:: 3.2.0
        """
        if not self:
            raise QueueEmpty("There are no items currently in this queue.")

        return self[index]

    def swap(self, first: int, second: int, /) -> None:
        """Swap two items in the queue by index.

        Parameters
        ----------
        first: int
            The first index to swap with.
        second: int
            The second index to swap with.

        Returns
        -------
        None

        Raises
        ------
        IndexError
            No track exists at the given index.

        Example
        -------

        .. code:: python3

            # Swap the first and second tracks in the queue.
            queue.swap(0, 1)


        .. versionadded:: 3.2.0
        """
        self[first], self[second] = self[second], self[first]

    def index(self, item: Playable, /) -> int:
        """Return the index of the first occurence of a :class:`wavelink.Playable` in the queue.

        Parameters
        ----------
        item: :class:`wavelink.Playable`
            The item to search the index for.

        Returns
        -------
        int
            The index of the item in the queue.

        Raises
        ------
        ValueError
            The item was not found in the queue.


        .. versionadded:: 3.2.0
        """
        return self._items.index(item)

    def shuffle(self) -> None:
        """Shuffles the queue in place. This does **not** return anything.

        Example
        -------

        .. code:: python3

            player.queue.shuffle()
            # Your queue has now been shuffled...

        Returns
        -------
        None
        """

        random.shuffle(self._items)

    def clear(self) -> None:
        """Remove all items from the queue.

        .. note::

            This does not reset the queue or clear history. Use this method on queue.history to clear history.

        Example
        -------
        .. code:: python3

            player.queue.clear()
            # Your queue is now empty...

        Returns
        -------
        None
        """

        self._items.clear()

    def copy(self) -> Queue:
        """Create a shallow copy of the queue.

        Returns
        -------
        :class:`wavelink.Queue`
            A shallow copy of the queue.
        """

        copy_queue = Queue(history=self.history is not None)
        copy_queue._items = self._items.copy()
        return copy_queue

    def reset(self) -> None:
        """Reset the queue to its default state. This will clear the queue and history.

        .. note::

            This will cancel any waiting futures on the queue. E.g. :meth:`wavelink.Queue.get_wait`.

        Returns
        -------
        None
        """
        self.clear()
        if self.history is not None:
            self.history.clear()

        for waiter in self._waiters:
            waiter.cancel()

        self._waiters.clear()

        self._mode: QueueMode = QueueMode.normal
        self._loaded = None

    def remove(self, item: Playable, /, count: int | None = 1) -> int:
        """Remove a specific track from the queue up to a given count or all instances.

        .. note::

            This method starts from the left hand side of the queue E.g. the beginning.

        .. warning::

            Setting count to ``<= 0`` is equivalent to setting it to ``1``.

        Parameters
        ----------
        item: :class:`wavelink.Playable`
            The item to remove from the queue.
        count: int
            The amount of times to remove the item from the queue. Defaults to ``1``.
            If set to ``None`` this will remove all instances of the item.

        Returns
        -------
        int
            The amount of times the item was removed from the queue.

        Raises
        ------
        ValueError
            The item was not found in the queue.


        .. versionadded:: 3.2.0
        """
        deleted_count: int = 0

        for track in self._items.copy():
            if track == item:
                self._items.remove(track)
                deleted_count += 1

                if count is not None and deleted_count >= count:
                    break

        return deleted_count

    @property
    def loaded(self) -> Playable | None:
        """The currently loaded track that will repeat when the queue is set to :attr:`wavelink.QueueMode.loop`.

        This track will be retrieved when using :meth:`wavelink.Queue.get` if the queue is in loop mode.
        You can unload the track by setting this property to ``None`` or by using :meth:`wavelink.Player.skip` with
        ``force=True``.

        Setting this property to a new :class:`wavelink.Playable` will replace the currently loaded track, but will not
        add it to the queue; or history until the track is played.

        Returns
        -------
        :class:`wavelink.Playable` | None
            The currently loaded track or ``None`` if there is no track ready to repeat.

        Raises
        ------
        TypeError
            The track was not a :class:`wavelink.Playable` or ``None``.


        .. versionadded:: 3.2.0
        """
        return self._loaded

    @loaded.setter
    def loaded(self, value: Playable | None) -> None:
        if value is not None:
            self._check_compatibility(value)

        self._loaded = value



"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import Any, TypedDict


class Equalizer(TypedDict):
    band: int
    gain: float


class Karaoke(TypedDict, total=False):
    level: float | None
    monoLevel: float | None
    filterBand: float | None
    filterWidth: float | None


class Timescale(TypedDict, total=False):
    speed: float | None
    pitch: float | None
    rate: float | None


class Tremolo(TypedDict, total=False):
    frequency: float | None
    depth: float | None


class Vibrato(TypedDict, total=False):
    frequency: float | None
    depth: float | None


class Rotation(TypedDict, total=False):
    rotationHz: float | None


class Distortion(TypedDict, total=False):
    sinOffset: float | None
    sinScale: float | None
    cosOffset: float | None
    cosScale: float | None
    tanOffset: float | None
    tanScale: float | None
    offset: float | None
    scale: float | None


class ChannelMix(TypedDict, total=False):
    leftToLeft: float | None
    leftToRight: float | None
    rightToLeft: float | None
    rightToRight: float | None


class LowPass(TypedDict, total=False):
    smoothing: float | None


class FilterPayload(TypedDict, total=False):
    volume: float | None
    equalizer: list[Equalizer] | None
    karaoke: Karaoke
    timescale: Timescale
    tremolo: Tremolo
    vibrato: Vibrato
    rotation: Rotation
    distortion: Distortion
    channelMix: ChannelMix
    lowPass: LowPass
    pluginFilters: dict[str, Any]



"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import asyncio
import logging
import random
import time
from collections import deque
from typing import TYPE_CHECKING, Any, TypeAlias

import async_timeout
import discord
from discord.abc import Connectable
from discord.utils import MISSING

import wavelink

from .enums import AutoPlayMode, NodeStatus, QueueMode
from .exceptions import (
    ChannelTimeoutException,
    InvalidChannelStateException,
    LavalinkException,
    LavalinkLoadException,
    QueueEmpty,
)
from .filters import Filters
from .node import Pool
from .payloads import (
    PlayerUpdateEventPayload,
    TrackEndEventPayload,
    TrackStartEventPayload,
)
from .queue import Queue
from .tracks import Playable, Playlist

if TYPE_CHECKING:
    from discord.types.voice import GuildVoiceState as GuildVoiceStatePayload
    from discord.types.voice import VoiceServerUpdate as VoiceServerUpdatePayload
    from typing_extensions import Self

    from .node import Node
    from .types.request import Request as RequestPayload
    from .types.state import PlayerVoiceState, VoiceState

    VocalGuildChannel = discord.VoiceChannel | discord.StageChannel

logger: logging.Logger = logging.getLogger(__name__)


T_a: TypeAlias = list[Playable] | Playlist


class Player(discord.VoiceProtocol):
    """The Player is a :class:`discord.VoiceProtocol` used to connect your :class:`discord.Client` to a
    :class:`discord.VoiceChannel`.

    The player controls the music elements of the bot including playing tracks, the queue, connecting etc.
    See Also: The various methods available.

    .. note::

        Since the Player is a :class:`discord.VoiceProtocol`, it is attached to the various ``voice_client`` attributes
        in discord.py, including ``guild.voice_client``, ``ctx.voice_client`` and ``interaction.voice_client``.

    Attributes
    ----------
    queue: :class:`~wavelink.Queue`
        The queue associated with this player.
    auto_queue: :class:`~wavelink.Queue`
        The auto_queue associated with this player. This queue holds tracks that are recommended by the AutoPlay feature.
    """

    channel: VocalGuildChannel

    def __call__(self, client: discord.Client, channel: VocalGuildChannel) -> Self:
        super().__init__(client, channel)

        self._guild = channel.guild

        return self

    def __init__(
        self, client: discord.Client = MISSING, channel: Connectable = MISSING, *, nodes: list[Node] | None = None
    ) -> None:
        super().__init__(client, channel)

        self.client: discord.Client = client
        self._guild: discord.Guild | None = None

        self._voice_state: PlayerVoiceState = {"voice": {}}

        self._node: Node
        if not nodes:
            self._node = Pool.get_node()
        else:
            self._node = sorted(nodes, key=lambda n: len(n.players))[0]

        if self.client is MISSING and self.node.client:
            self.client = self.node.client

        self._last_update: int | None = None
        self._last_position: int = 0
        self._ping: int = -1

        self._connected: bool = False
        self._connection_event: asyncio.Event = asyncio.Event()

        self._current: Playable | None = None
        self._original: Playable | None = None
        self._previous: Playable | None = None

        self.queue: Queue = Queue()
        self.auto_queue: Queue = Queue()

        self._volume: int = 100
        self._paused: bool = False

        self._auto_cutoff: int = 20
        self._auto_weight: int = 3
        self._previous_seeds_cutoff: int = self._auto_cutoff * self._auto_weight
        self._history_count: int | None = None

        self._autoplay: AutoPlayMode = AutoPlayMode.disabled
        self.__previous_seeds: asyncio.Queue[str] = asyncio.Queue(maxsize=self._previous_seeds_cutoff)

        self._auto_lock: asyncio.Lock = asyncio.Lock()
        self._error_count: int = 0

        self._filters: Filters = Filters()

        # Needed for the inactivity checks...
        self._inactivity_task: asyncio.Task[bool] | None = None
        self._inactivity_wait: int | None = self._node._inactive_player_timeout

    def _inactivity_task_callback(self, task: asyncio.Task[bool]) -> None:
        result: bool = task.result()
        cancelled: bool = task.cancelled()

        if cancelled or result is False:
            logger.debug("Disregarding Inactivity Check Task <%s> as it was previously cancelled.", task.get_name())
            return

        if result is not True:
            logger.debug("Disregarding Inactivity Check Task <%s> as it received an unknown result.", task.get_name())
            return

        if not self._guild:
            logger.debug("Disregarding Inactivity Check Task <%s> as it has no guild.", task.get_name())
            return

        if self.playing:
            logger.debug(
                "Disregarding Inactivity Check Task <%s> as Player <%s> is playing.", task.get_name(), self._guild.id
            )
            return

        self.client.dispatch("wavelink_inactive_player", self)
        logger.debug('Dispatched "on_wavelink_inactive_player" for Player <%s>.', self._guild.id)

    async def _inactivity_runner(self, wait: int) -> bool:
        try:
            await asyncio.sleep(wait)
        except asyncio.CancelledError:
            return False

        return True

    def _inactivity_cancel(self) -> None:
        if self._inactivity_task:
            try:
                self._inactivity_task.cancel()
            except Exception:
                pass

        self._inactivity_task = None

    def _inactivity_start(self) -> None:
        if self._inactivity_wait is not None and self._inactivity_wait > 0:
            self._inactivity_task = asyncio.create_task(self._inactivity_runner(self._inactivity_wait))
            self._inactivity_task.add_done_callback(self._inactivity_task_callback)

    async def _track_start(self, payload: TrackStartEventPayload) -> None:
        self._inactivity_cancel()

    async def _auto_play_event(self, payload: TrackEndEventPayload) -> None:
        if self._autoplay is AutoPlayMode.disabled:
            self._inactivity_start()
            return

        if self._error_count >= 3:
            logger.warning(
                "AutoPlay was unable to continue as you have received too many consecutive errors."
                "Please check the error log on Lavalink."
            )
            self._inactivity_start()
            return

        if payload.reason == "replaced":
            self._error_count = 0
            return

        elif payload.reason == "loadFailed":
            self._error_count += 1

        else:
            self._error_count = 0

        if self.node.status is not NodeStatus.CONNECTED:
            logger.warning(f'"Unable to use AutoPlay on Player for Guild "{self.guild}" due to disconnected Node.')
            return

        if not isinstance(self.queue, Queue) or not isinstance(self.auto_queue, Queue):  # type: ignore
            logger.warning(f'"Unable to use AutoPlay on Player for Guild "{self.guild}" due to unsupported Queue.')
            self._inactivity_start()
            return

        if self.queue.mode is QueueMode.loop:
            await self._do_partial(history=False)

        elif self.queue.mode is QueueMode.loop_all:
            await self._do_partial()

        elif self._autoplay is AutoPlayMode.partial or self.queue:
            await self._do_partial()

        elif self._autoplay is AutoPlayMode.enabled:
            async with self._auto_lock:
                await self._do_recommendation()

    async def _do_partial(self, *, history: bool = True) -> None:
        # We still do the inactivity start here since if play fails and we have no more tracks...
        # we should eventually fire the inactivity event...
        self._inactivity_start()

        if self._current is None:
            try:
                track: Playable = self.queue.get()
            except QueueEmpty:
                return

            await self.play(track, add_history=history)

    async def _do_recommendation(self):
        assert self.guild is not None
        assert self.queue.history is not None and self.auto_queue.history is not None

        if len(self.auto_queue) > self._auto_cutoff + 1:
            # We still do the inactivity start here since if play fails and we have no more tracks...
            # we should eventually fire the inactivity event...
            self._inactivity_start()

            track: Playable = self.auto_queue.get()
            self.auto_queue.history.put(track)

            await self.play(track, add_history=False)
            return

        weighted_history: list[Playable] = self.queue.history[::-1][: max(5, 5 * self._auto_weight)]
        weighted_upcoming: list[Playable] = self.auto_queue[: max(3, int((5 * self._auto_weight) / 3))]
        choices: list[Playable | None] = [*weighted_history, *weighted_upcoming, self._current, self._previous]

        # Filter out tracks which are None...
        _previous: deque[str] = self.__previous_seeds._queue  # type: ignore
        seeds: list[Playable] = [t for t in choices if t is not None and t.identifier not in _previous]
        random.shuffle(seeds)

        spotify: list[str] = [t.identifier for t in seeds if t.source == "spotify"]
        youtube: list[str] = [t.identifier for t in seeds if t.source == "youtube"]

        spotify_query: str | None = None
        youtube_query: str | None = None

        count: int = len(self.queue.history)
        changed_by: int = min(3, count) if self._history_count is None else count - self._history_count

        if changed_by > 0:
            self._history_count = count

        changed_history: list[Playable] = self.queue.history[::-1]

        added: int = 0
        for i in range(min(changed_by, 3)):
            track: Playable = changed_history[i]

            if added == 2 and track.source == "spotify":
                break

            if track.source == "spotify":
                spotify.insert(0, track.identifier)
                added += 1

            elif track.source == "youtube":
                youtube[0] = track.identifier

        if spotify:
            spotify_seeds: list[str] = spotify[:3]
            spotify_query = f"sprec:seed_tracks={','.join(spotify_seeds)}&limit=10"

            for s_seed in spotify_seeds:
                self._add_to_previous_seeds(s_seed)

        if youtube:
            ytm_seed: str = youtube[0]
            youtube_query = f"https://music.youtube.com/watch?v={ytm_seed}8&list=RD{ytm_seed}"
            self._add_to_previous_seeds(ytm_seed)

        async def _search(query: str | None) -> T_a:
            if query is None:
                return []

            try:
                search: wavelink.Search = await Pool.fetch_tracks(query)
            except (LavalinkLoadException, LavalinkException):
                return []

            if not search:
                return []

            tracks: list[Playable]
            if isinstance(search, Playlist):
                tracks = search.tracks.copy()
            else:
                tracks = search

            return tracks

        results: tuple[T_a, T_a] = await asyncio.gather(_search(spotify_query), _search(youtube_query))

        # track for result in results for track in result...
        filtered_r: list[Playable] = [t for r in results for t in r]

        if not filtered_r:
            logger.debug(f'Player "{self.guild.id}" could not load any songs via AutoPlay.')
            self._inactivity_start()
            return

        if not self._current:
            now: Playable = filtered_r.pop(1)
            now._recommended = True
            self.auto_queue.history.put(now)

            await self.play(now, add_history=False)

        # Possibly adjust these thresholds?
        history: list[Playable] = (
            self.auto_queue[:40] + self.queue[:40] + self.queue.history[:-41:-1] + self.auto_queue.history[:-61:-1]
        )

        added: int = 0
        for track in filtered_r:
            if track in history:
                continue

            track._recommended = True
            added += await self.auto_queue.put_wait(track)

        random.shuffle(self.auto_queue._items)
        logger.debug(f'Player "{self.guild.id}" added "{added}" tracks to the auto_queue via AutoPlay.')

        # Probably don't need this here as it's likely to be cancelled instantly...
        self._inactivity_start()

    @property
    def inactive_timeout(self) -> int | None:
        """A property which returns the time as an ``int`` of seconds to wait before this player dispatches the
        :func:`on_wavelink_inactive_player` event.

        This property could return ``None`` if no time has been set.

        An inactive player is a player that has not been playing anything for the specified amount of seconds.

        - Pausing the player while a song is playing will not activate this countdown.
        - The countdown starts when a track ends and cancels when a track starts.
        - The countdown will not trigger until a song is played for the first time or this property is reset.
        - The default countdown for all players is set on :class:`~wavelink.Node`.

        This property can be set with a valid ``int`` of seconds to wait before dispatching the
        :func:`on_wavelink_inactive_player` event or ``None`` to remove the timeout.


        .. warning::

            Setting this to a value of ``0`` or below is the equivalent of setting this property to ``None``.


        When this property is set, the timeout will reset, and all previously waiting countdowns are cancelled.

        - See: :class:`~wavelink.Node`
        - See: :func:`on_wavelink_inactive_player`


        .. versionadded:: 3.2.0
        """
        return self._inactivity_wait

    @inactive_timeout.setter
    def inactive_timeout(self, value: int | None) -> None:
        if not value or value <= 0:
            self._inactivity_wait = None
            self._inactivity_cancel()
            return

        if value < 10:
            logger.warn('Setting "inactive_timeout" below 10 seconds may result in unwanted side effects.')

        self._inactivity_wait = value
        self._inactivity_cancel()

        if self.connected and not self.playing:
            self._inactivity_start()

    @property
    def autoplay(self) -> AutoPlayMode:
        """A property which returns the :class:`wavelink.AutoPlayMode` the player is currently in.

        This property can be set with any :class:`wavelink.AutoPlayMode` enum value.


        .. versionchanged:: 3.0.0

            This property now accepts and returns a :class:`wavelink.AutoPlayMode` enum value.
        """
        return self._autoplay

    @autoplay.setter
    def autoplay(self, value: Any) -> None:
        if not isinstance(value, AutoPlayMode):
            raise ValueError("Please provide a valid 'wavelink.AutoPlayMode' to set.")

        self._autoplay = value

    @property
    def node(self) -> Node:
        """The :class:`Player`'s currently selected :class:`Node`.


        .. versionchanged:: 3.0.0

            This property was previously known as ``current_node``.
        """
        return self._node

    @property
    def guild(self) -> discord.Guild | None:
        """Returns the :class:`Player`'s associated :class:`discord.Guild`.

        Could be None if this :class:`Player` has not been connected.
        """
        return self._guild

    @property
    def connected(self) -> bool:
        """Returns a bool indicating if the player is currently connected to a voice channel.

        .. versionchanged:: 3.0.0

            This property was previously known as ``is_connected``.
        """
        return self.channel and self._connected

    @property
    def current(self) -> Playable | None:
        """Returns the currently playing :class:`~wavelink.Playable` or None if no track is playing."""
        return self._current

    @property
    def volume(self) -> int:
        """Returns an int representing the currently set volume, as a percentage.

        See: :meth:`set_volume` for setting the volume.
        """
        return self._volume

    @property
    def filters(self) -> Filters:
        """Property which returns the :class:`~wavelink.Filters` currently assigned to the Player.

        See: :meth:`~wavelink.Player.set_filters` for setting the players filters.

        .. versionchanged:: 3.0.0

            This property was previously known as ``filter``.
        """
        return self._filters

    @property
    def paused(self) -> bool:
        """Returns the paused status of the player. A currently paused player will return ``True``.

        See: :meth:`pause` and :meth:`play` for setting the paused status.
        """
        return self._paused

    @property
    def ping(self) -> int:
        """Returns the ping in milliseconds as int between your connected Lavalink Node and Discord (Players Channel).

        Returns ``-1`` if no player update event has been received or the player is not connected.
        """
        return self._ping

    @property
    def playing(self) -> bool:
        """Returns whether the :class:`~Player` is currently playing a track and is connected.

        Due to relying on validation from Lavalink, this property may in some cases return ``True`` directly after
        skipping/stopping a track, although this is not the case when disconnecting the player.

        This property will return ``True`` in cases where the player is paused *and* has a track loaded.

        .. versionchanged:: 3.0.0

            This property used to be known as the `is_playing()` method.
        """
        return self._connected and self._current is not None

    @property
    def position(self) -> int:
        """Returns the position of the currently playing :class:`~wavelink.Playable` in milliseconds.

        This property relies on information updates from Lavalink.

        In cases there is no :class:`~wavelink.Playable` loaded or the player is not connected,
        this property will return ``0``.

        This property will return ``0`` if no update has been received from Lavalink.

        .. versionchanged:: 3.0.0

            This property now uses a monotonic clock.
        """
        if self.current is None or not self.playing:
            return 0

        if not self.connected:
            return 0

        if self._last_update is None:
            return 0

        if self.paused:
            return self._last_position

        position: int = int((time.monotonic_ns() - self._last_update) / 1000000) + self._last_position
        return min(position, self.current.length)

    async def _update_event(self, payload: PlayerUpdateEventPayload) -> None:
        # Convert nanoseconds into milliseconds...
        self._last_update = time.monotonic_ns()
        self._last_position = payload.position

        self._ping = payload.ping

    async def on_voice_state_update(self, data: GuildVoiceStatePayload, /) -> None:
        channel_id = data["channel_id"]

        if not channel_id:
            await self._destroy()
            return

        self._connected = True

        self._voice_state["voice"]["session_id"] = data["session_id"]
        self.channel = self.client.get_channel(int(channel_id))  # type: ignore

    async def on_voice_server_update(self, data: VoiceServerUpdatePayload, /) -> None:
        self._voice_state["voice"]["token"] = data["token"]
        self._voice_state["voice"]["endpoint"] = data["endpoint"]

        await self._dispatch_voice_update()

    async def _dispatch_voice_update(self) -> None:
        assert self.guild is not None
        data: VoiceState = self._voice_state["voice"]

        try:
            session_id: str = data["session_id"]
            token: str = data["token"]
        except KeyError:
            return

        endpoint: str | None = data.get("endpoint", None)
        if not endpoint:
            return

        request: RequestPayload = {"voice": {"sessionId": session_id, "token": token, "endpoint": endpoint}}

        try:
            await self.node._update_player(self.guild.id, data=request)
        except LavalinkException:
            await self.disconnect()
        else:
            self._connection_event.set()

        logger.debug(f"Player {self.guild.id} is dispatching VOICE_UPDATE.")

    async def connect(
        self, *, timeout: float = 10.0, reconnect: bool, self_deaf: bool = False, self_mute: bool = False
    ) -> None:
        """

        .. warning::

            Do not use this method directly on the player. See: :meth:`discord.VoiceChannel.connect` for more details.


        Pass the :class:`wavelink.Player` to ``cls=`` in :meth:`discord.VoiceChannel.connect`.


        Raises
        ------
        ChannelTimeoutException
            Connecting to the voice channel timed out.
        InvalidChannelStateException
            You tried to connect this player without an appropriate voice channel.
        """
        if self.channel is MISSING:
            msg: str = 'Please use "discord.VoiceChannel.connect(cls=...)" and pass this Player to cls.'
            raise InvalidChannelStateException(f"Player tried to connect without a valid channel: {msg}")

        if not self._guild:
            self._guild = self.channel.guild

        self.node._players[self._guild.id] = self

        assert self.guild is not None
        await self.guild.change_voice_state(channel=self.channel, self_mute=self_mute, self_deaf=self_deaf)

        try:
            async with async_timeout.timeout(timeout):
                await self._connection_event.wait()
        except (asyncio.TimeoutError, asyncio.CancelledError):
            msg = f"Unable to connect to {self.channel} as it exceeded the timeout of {timeout} seconds."
            raise ChannelTimeoutException(msg)

    async def move_to(
        self,
        channel: VocalGuildChannel | None,
        *,
        timeout: float = 10.0,
        self_deaf: bool | None = None,
        self_mute: bool | None = None,
    ) -> None:
        """Method to move the player to another channel.

        Parameters
        ----------
        channel: :class:`discord.VoiceChannel` | :class:`discord.StageChannel`
            The new channel to move to.
        timeout: float
            The timeout in ``seconds`` before raising. Defaults to 10.0.
        self_deaf: bool | None
            Whether to deafen when moving. Defaults to ``None`` which keeps the current setting or ``False``
            if they can not be determined.
        self_mute: bool | None
            Whether to self mute when moving. Defaults to ``None`` which keeps the current setting or ``False``
            if they can not be determined.

        Raises
        ------
        ChannelTimeoutException
            Connecting to the voice channel timed out.
        InvalidChannelStateException
            You tried to connect this player without an appropriate guild.
        """
        if not self.guild:
            raise InvalidChannelStateException("Player tried to move without a valid guild.")

        self._connection_event.clear()
        voice: discord.VoiceState | None = self.guild.me.voice

        if self_deaf is None and voice:
            self_deaf = voice.self_deaf

        if self_mute is None and voice:
            self_mute = voice.self_mute

        self_deaf = bool(self_deaf)
        self_mute = bool(self_mute)

        await self.guild.change_voice_state(channel=channel, self_mute=self_mute, self_deaf=self_deaf)

        if channel is None:
            return

        try:
            async with async_timeout.timeout(timeout):
                await self._connection_event.wait()
        except (asyncio.TimeoutError, asyncio.CancelledError):
            msg = f"Unable to connect to {channel} as it exceeded the timeout of {timeout} seconds."
            raise ChannelTimeoutException(msg)

    async def play(
        self,
        track: Playable,
        *,
        replace: bool = True,
        start: int = 0,
        end: int | None = None,
        volume: int | None = None,
        paused: bool | None = None,
        add_history: bool = True,
        filters: Filters | None = None,
    ) -> Playable:
        """Play the provided :class:`~wavelink.Playable`.

        Parameters
        ----------
        track: :class:`~wavelink.Playable`
            The track to being playing.
        replace: bool
            Whether this track should replace the currently playing track, if there is one. Defaults to ``True``.
        start: int
            The position to start playing the track at in milliseconds.
            Defaults to ``0`` which will start the track from the beginning.
        end: Optional[int]
            The position to end the track at in milliseconds.
            Defaults to ``None`` which means this track will play until the very end.
        volume: Optional[int]
            Sets the volume of the player. Must be between ``0`` and ``1000``.
            Defaults to ``None`` which will not change the current volume.
            See Also: :meth:`set_volume`
        paused: bool | None
            Whether the player should be paused, resumed or retain current status when playing this track.
            Setting this parameter to ``True`` will pause the player. Setting this parameter to ``False`` will
            resume the player if it is currently paused. Setting this parameter to ``None`` will not change the status
            of the player. Defaults to ``None``.
        add_history: Optional[bool]
            If this argument is set to ``True``, the :class:`~Player` will add this track into the
            :class:`wavelink.Queue` history, if loading the track was successful. If ``False`` this track will not be
            added to your history. This does not directly affect the ``AutoPlay Queue`` but will alter how ``AutoPlay``
            recommends songs in the future. Defaults to ``True``.
        filters: Optional[:class:`~wavelink.Filters`]
            An Optional[:class:`~wavelink.Filters`] to apply when playing this track. Defaults to ``None``.
            If this is ``None`` the currently set filters on the player will be applied.


        Returns
        -------
        :class:`~wavelink.Playable`
            The track that began playing.


        .. versionchanged:: 3.0.0

            Added the ``paused`` parameter. Parameters ``replace``, ``start``, ``end``, ``volume`` and ``paused``
            are now all keyword-only arguments.

            Added the ``add_history`` keyword-only argument.

            Added the ``filters`` keyword-only argument.
        """
        assert self.guild is not None

        original_vol: int = self._volume
        vol: int = volume or self._volume

        if vol != self._volume:
            self._volume = vol

        if replace or not self._current:
            self._current = track
            self._original = track

        old_previous = self._previous
        self._previous = self._current

        pause: bool
        if paused is not None:
            pause = paused
        else:
            pause = self._paused

        if filters:
            self._filters = filters

        request: RequestPayload = {
            "track": {"encoded": track.encoded, "userData": dict(track.extras)},
            "volume": vol,
            "position": start,
            "endTime": end,
            "paused": pause,
            "filters": self._filters(),
        }

        try:
            await self.node._update_player(self.guild.id, data=request, replace=replace)
        except LavalinkException as e:
            self._current = None
            self._original = None
            self._previous = old_previous
            self._volume = original_vol
            raise e

        self._paused = pause

        if add_history:
            assert self.queue.history is not None
            self.queue.history.put(track)

        return track

    async def pause(self, value: bool, /) -> None:
        """Set the paused or resume state of the player.

        Parameters
        ----------
        value: bool
            A bool indicating whether the player should be paused or resumed. True indicates that the player should be
            ``paused``. False will resume the player if it is currently paused.


        .. versionchanged:: 3.0.0

            This method now expects a positional-only bool value. The ``resume`` method has been removed.
        """
        assert self.guild is not None

        request: RequestPayload = {"paused": value}
        await self.node._update_player(self.guild.id, data=request)

        self._paused = value

    async def seek(self, position: int = 0, /) -> None:
        """Seek to the provided position in the currently playing track, in milliseconds.

        Parameters
        ----------
        position: int
            The position to seek to in milliseconds. To restart the song from the beginning,
            you can disregard this parameter or set position to 0.


        .. versionchanged:: 3.0.0

            The ``position`` parameter is now positional-only, and has a default of 0.
        """
        assert self.guild is not None

        if not self._current:
            return

        request: RequestPayload = {"position": position}
        await self.node._update_player(self.guild.id, data=request)

    async def set_filters(self, filters: Filters | None = None, /, *, seek: bool = False) -> None:
        """Set the :class:`wavelink.Filters` on the player.

        Parameters
        ----------
        filters: Optional[:class:`~wavelink.Filters`]
            The filters to set on the player. Could be ``None`` to reset the currently applied filters.
            Defaults to ``None``.
        seek: bool
            Whether to seek immediately when applying these filters. Seeking uses more resources, but applies the
            filters immediately. Defaults to ``False``.


        .. versionchanged:: 3.0.0

            This method now accepts a positional-only argument of filters, which now defaults to None. Filters
            were redesigned in this version, see: :class:`wavelink.Filters`.


        .. versionchanged:: 3.0.0

            This method was previously known as ``set_filter``.
        """
        assert self.guild is not None

        if filters is None:
            filters = Filters()

        request: RequestPayload = {"filters": filters()}
        await self.node._update_player(self.guild.id, data=request)
        self._filters = filters

        if self.playing and seek:
            await self.seek(self.position)

    async def set_volume(self, value: int = 100, /) -> None:
        """Set the :class:`Player` volume, as a percentage, between 0 and 1000.

        By default, every player is set to 100 on creation. If a value outside 0 to 1000 is provided it will be
        clamped.

        Parameters
        ----------
        value: int
            A volume value between 0 and 1000. To reset the player to 100, you can disregard this parameter.


        .. versionchanged:: 3.0.0

            The ``value`` parameter is now positional-only, and has a default of 100.
        """
        assert self.guild is not None
        vol: int = max(min(value, 1000), 0)

        request: RequestPayload = {"volume": vol}
        await self.node._update_player(self.guild.id, data=request)

        self._volume = vol

    async def disconnect(self, **kwargs: Any) -> None:
        """Disconnect the player from the current voice channel and remove it from the :class:`~wavelink.Node`.

        This method will cause any playing track to stop and potentially trigger the following events:

            - ``on_wavelink_track_end``
            - ``on_wavelink_websocket_closed``


        .. warning::

            Please do not re-use a :class:`Player` instance that has been disconnected, unwanted side effects are
            possible.
        """
        assert self.guild

        await self._destroy()
        await self.guild.change_voice_state(channel=None)

    async def stop(self, *, force: bool = True) -> Playable | None:
        """An alias to :meth:`skip`.

        See Also: :meth:`skip` for more information.

        .. versionchanged:: 3.0.0

            This method is now known as ``skip``, but the alias ``stop`` has been kept for backwards compatibility.
        """
        return await self.skip(force=force)

    async def skip(self, *, force: bool = True) -> Playable | None:
        """Stop playing the currently playing track.

        Parameters
        ----------
        force: bool
            Whether the track should skip looping, if :class:`wavelink.Queue` has been set to loop.
            Defaults to ``True``.

        Returns
        -------
        :class:`~wavelink.Playable` | None
            The currently playing track that was skipped, or ``None`` if no track was playing.


        .. versionchanged:: 3.0.0

            This method was previously known as ``stop``. To avoid confusion this method is now known as ``skip``.
            This method now returns the :class:`~wavelink.Playable` that was skipped.
        """
        assert self.guild is not None
        old: Playable | None = self._current

        if force:
            self.queue._loaded = None

        request: RequestPayload = {"track": {"encoded": None}}
        await self.node._update_player(self.guild.id, data=request, replace=True)

        return old

    def _invalidate(self) -> None:
        self._connected = False
        self._connection_event.clear()
        self._inactivity_cancel()

        try:
            self.cleanup()
        except (AttributeError, KeyError):
            pass

    async def _destroy(self) -> None:
        assert self.guild

        self._invalidate()
        player: Player | None = self.node._players.pop(self.guild.id, None)

        if player:
            try:
                await self.node._destroy_player(self.guild.id)
            except LavalinkException:
                pass

    def _add_to_previous_seeds(self, seed: str) -> None:
        # Helper method to manage previous seeds.
        if self.__previous_seeds.full():
            self.__previous_seeds.get_nowait()
        self.__previous_seeds.put_nowait(seed)



"""
MIT License

Copyright (c) 2019-Current PythonistaGuild, EvieePy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Any

from .exceptions import WavelinkException


class CapacityZero(WavelinkException):
    ...


class _MissingSentinel:
    __slots__ = ()

    def __eq__(self, other: object) -> bool:
        return False

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:
        return 0

    def __repr__(self) -> str:
        return "..."


class _NotFoundSentinel(_MissingSentinel):
    def __repr__(self) -> str:
        return "NotFound"


MISSING: Any = _MissingSentinel()
NotFound: Any = _NotFoundSentinel()


class DLLNode:
    __slots__ = ("value", "previous", "later")

    def __init__(self, value: Any | None = None, previous: DLLNode | None = None, later: DLLNode | None = None) -> None:
        self.value = value
        self.previous = previous
        self.later = later


@dataclass
class DataNode:
    key: Any
    value: Any
    frequency: int
    node: DLLNode


class LFUCache:
    def __init__(self, *, capacity: int) -> None:
        self._capacity = capacity
        self._cache: dict[Any, DataNode] = {}

        self._freq_map: defaultdict[int, DLL] = defaultdict(DLL)
        self._min: int = 1
        self._used: int = 0

    def __len__(self) -> int:
        return len(self._cache)

    def __getitem__(self, key: Any) -> Any:
        if key not in self._cache:
            raise KeyError(f'"{key}" could not be found in LFU.')

        return self.get(key)

    def __setitem__(self, key: Any, value: Any) -> None:
        return self.put(key, value)

    @property
    def capacity(self) -> int:
        return self._capacity

    def get(self, key: Any, default: Any = MISSING) -> Any:
        if key not in self._cache:
            return default if default is not MISSING else NotFound

        data: DataNode = self._cache[key]
        self._freq_map[data.frequency].remove(data.node)
        self._freq_map[data.frequency + 1].append(data.node)

        self._cache[key] = DataNode(key=key, value=data.value, frequency=data.frequency + 1, node=data.node)
        self._min += self._min == data.frequency and not self._freq_map[data.frequency]

        return data.value

    def put(self, key: Any, value: Any) -> None:
        if self._capacity <= 0:
            raise CapacityZero("Unable to place item in LFU as capacity has been set to 0 or below.")

        if key in self._cache:
            self._cache[key].value = value
            self.get(key)
            return

        if self._used == self._capacity:
            least_freq: DLL = self._freq_map[self._min]
            least_freq_key: DLLNode | None = least_freq.popleft()

            if least_freq_key:
                self._cache.pop(least_freq_key.value)
                self._used -= 1

        data: DataNode = DataNode(key=key, value=value, frequency=1, node=DLLNode(key))
        self._freq_map[data.frequency].append(data.node)
        self._cache[key] = data

        self._used += 1
        self._min = 1


class DLL:
    __slots__ = ("head", "tail")

    def __init__(self) -> None:
        self.head: DLLNode = DLLNode()
        self.tail: DLLNode = DLLNode()

        self.head.later, self.tail.previous = self.tail, self.head

    def append(self, node: DLLNode) -> None:
        tail_prev: DLLNode | None = self.tail.previous
        tail: DLLNode | None = self.tail

        assert tail_prev and tail

        tail_prev.later = node
        tail.previous = node

        node.later = tail
        node.previous = tail_prev

    def popleft(self) -> DLLNode | None:
        node: DLLNode | None = self.head.later
        if node is None:
            return

        self.remove(node)
        return node

    def remove(self, node: DLLNode | None) -> None:
        if node is None:
            return

        node_prev: DLLNode | None = node.previous
        node_later: DLLNode | None = node.later

        assert node_prev and node_later

        node_prev.later = node_later
        node_later.previous = node_prev

        node.later = None
        node.previous = None

    def __bool__(self) -> bool:
        return self.head.later != self.tail
